import uuid
from typing import List, Optional

from fastapi import APIRouter, UploadFile, File, Body, HTTPException

from ... import Runtime400Exception
from ...models.status import StoreStatus, StoreItemStatus
from ...stores import workspace_store as store

router = APIRouter(prefix='/workspaces', tags=['workspaces'])


@router.get(
    path='', summary='Get all existing Workspaces\' status', response_model=StoreStatus
)
async def _get_items():
    return store.status


@router.delete(
    path='',
    summary='Deleting all Workspaces',
)
async def _clear_all():
    store.clear()


@router.delete(
    path='/{id}',
    summary='Deleting an existing Workspace',
)
async def _delete(id: uuid.UUID):
    try:
        store.delete(id=id, everything=True)
    except KeyError:
        raise HTTPException(status_code=404, detail=f'{id} not found in {store!r}')


@router.post(
    path='',
    summary='Upload files to a workspace',
    description='Return a UUID to the workspace, which can be used later when create Pea/Pod/Flow',
    response_model=uuid.UUID,
    status_code=201,
)
async def _create(
    files: List[UploadFile] = File(...), workspace_id: Optional[uuid.UUID] = Body(None)
):
    try:
        return store.add(files, workspace_id)
    except Exception as ex:
        raise Runtime400Exception from ex


@router.get(
    path='/{id}',
    summary='Get the status of an existing Workspace',
    response_model=StoreItemStatus,
)
async def _list(id: uuid.UUID):
    try:
        return store[id]
    except KeyError:
        raise HTTPException(status_code=404, detail=f'{id} not found in {store!r}')


@router.on_event('shutdown')
def _shutdown():
    store.reset()



import asyncio
import json
import uuid
from pathlib import Path
from typing import List

from fastapi import APIRouter, WebSocket, WebSocketDisconnect
from fastapi.exceptions import HTTPException
from fastapi.responses import FileResponse
from starlette.endpoints import WebSocketEndpoint
from starlette.types import Receive, Scope, Send
from websockets import ConnectionClosedOK
from websockets.exceptions import ConnectionClosedError

from ... import daemon_logger, jinad_args
from ...stores.helper import get_workspace_path

router = APIRouter(tags=['logs'])


@router.get(path='/logs/{workspace_id}/{log_id}')
async def _export_logs(workspace_id: uuid.UUID, log_id: uuid.UUID):
    filepath = get_workspace_path(workspace_id, log_id, 'logging.log')
    if not Path(filepath).is_file():
        raise HTTPException(
            status_code=404,
            detail=f'log file {log_id} not found in workspace {workspace_id}',
        )
    else:
        return FileResponse(filepath)


def _websocket_details(websocket: WebSocket):
    return f'{websocket.client.host}:{websocket.client.port}'


class ConnectionManager:
    """
    Manager of websockets listening for a log stream.

    TODO for now contian a single connection. Ideally there must be one
    manager per log with a thread checking for updates in log and broadcasting
    to active connections
    """

    def __init__(self):
        """Instantiate a ConnectionManager."""
        self.active_connections: List[WebSocket] = []

    async def connect(self, websocket: WebSocket):
        """
        Register a new websocket.

        :param websocket: websocket to register
        """
        await websocket.accept()
        daemon_logger.info(
            '%s is connected to stream logs!' % _websocket_details(websocket)
        )
        self.active_connections.append(websocket)

    async def disconnect(self, websocket: WebSocket):
        """
        Disconnect a websocket.

        :param websocket: websocket to disconnect
        """
        self.active_connections.remove(websocket)
        await websocket.close()
        daemon_logger.info('%s is disconnected' % _websocket_details(websocket))

    async def broadcast(self, message: dict):
        """
        Send a json message to all registered websockets.

        :param message: JSON-serializable message to be broadcast
        """
        daemon_logger.debug('connections: %r', self.active_connections)
        for connection in self.active_connections:
            try:
                await connection.send_json(message)
            except ConnectionClosedOK:
                pass
            except ConnectionClosedError:
                await self.disconnect(connection)


@router.websocket('/logstream/{workspace_id}/{log_id}')
async def _logstream(
    websocket: WebSocket, workspace_id: uuid.UUID, log_id: uuid.UUID, timeout: int = 0
):
    manager = ConnectionManager()
    await manager.connect(websocket)
    client_details = _websocket_details(websocket)
    filepath = get_workspace_path(workspace_id, log_id, 'logging.log')
    try:
        if jinad_args.no_fluentd:
            daemon_logger.warning(
                f'{client_details} asks for logstreaming but fluentd is not available'
            )
            return

        # on connection the fluentd file may not flushed (aka exist) yet
        n = 0
        while not Path(filepath).is_file():
            daemon_logger.debug(f'still waiting {filepath} to be ready...')
            await asyncio.sleep(1)
            n += 1
            if timeout > 0 and n >= timeout:
                return

        daemon_logger.success(f'{filepath} is ready for streaming')

        with open(filepath) as fp:
            fp.seek(0, 2)
            delay = 0.1
            n = 0
            while True:
                line = fp.readline()  # also possible to read an empty line
                if line:
                    daemon_logger.debug('sending line %s', line)
                    payload = None
                    try:
                        payload = json.loads(line)
                    except json.decoder.JSONDecodeError:
                        daemon_logger.warning(f'JSON decode error on {line}')

                    if payload:
                        await manager.broadcast(payload)
                        n = 0
                else:
                    await asyncio.sleep(delay)
                    n += 1
                    if timeout > 0 and n >= timeout / delay:
                        return
    except WebSocketDisconnect:
        await manager.disconnect(websocket)
    finally:
        await manager.disconnect(websocket)



"""A module for the websockets-based Client for Jina."""
import asyncio
from abc import ABC
from contextlib import nullcontext
from typing import Callable, Optional

from .base import BaseClient, InputType
from .grpc import GRPCClient
from .helper import callback_exec
from ..importer import ImportExtensions
from ..logging.profile import TimeContext, ProgressBar
from ..types.request import Request


class WebSocketClientMixin(BaseClient, ABC):
    """A MixIn for Websocket Client."""

    async def _get_results(
        self,
        inputs: InputType,
        on_done: Callable,
        on_error: Optional[Callable] = None,
        on_always: Optional[Callable] = None,
        **kwargs,
    ):
        """
        :meth:`send_requests()`
            Traverses through the request iterator
            Sends each request & awaits :meth:`websocket.send()`
            Sends & awaits `byte(True)` to acknowledge request iterator is empty
        Traversal logic:
            Starts an independent task :meth:`send_requests()`
            Awaits on each response from :meth:`websocket.recv()` (done in an async loop)
            This makes sure client makes concurrent invocations
        Await exit strategy:
            :meth:`send_requests()` keeps track of num_requests sent
            Async recv loop keeps track of num_responses received
            Client exits out of await when num_requests == num_responses

        :param inputs: the callable
        :param on_done: the callback for on_done
        :param on_error: the callback for on_error
        :param on_always: the callback for on_always
        :param kwargs: kwargs for _get_task_name and _get_requests
        :yields: generator over results
        """
        with ImportExtensions(required=True):
            import websockets

        self.inputs = inputs

        req_iter = self._get_requests(**kwargs)
        try:
            # setting `max_size` as None to avoid connection closure due to size of message
            # https://websockets.readthedocs.io/en/stable/api.html?highlight=1009#module-websockets.protocol

            async with websockets.connect(
                f'ws://{self.args.host}:{self.args.port_expose}/',
                max_size=None,
                ping_interval=None,
            ) as websocket:
                # To enable websockets debug logs
                # https://websockets.readthedocs.io/en/stable/cheatsheet.html#debugging
                self.logger.success(
                    f'connected to the gateway at {self.args.host}:{self.args.port_expose}!'
                )
                self.num_requests = 0
                self.num_responses = 0

                async def _send_requests(request_iterator):
                    next_request = None
                    for next_request in request_iterator:
                        await websocket.send(next_request.SerializeToString())
                        self.num_requests += 1
                    # Check if there was any request generated
                    if next_request is not None:
                        # Server has no way of knowing when to stop the await on sending response back to the client
                        # We send one last message to say `request_iterator` is completed.
                        # On the client side, this :meth:`send` doesn't need to be awaited with a :meth:`recv`
                        await websocket.send(bytes(True))
                    else:
                        # There is nothing to send, disconnect gracefully
                        await websocket.close(reason='No data to send')

                if self.args.show_progress:
                    cm1, cm2 = ProgressBar(), TimeContext('')
                else:
                    cm1, cm2 = nullcontext(), nullcontext()

                with cm1 as p_bar, cm2:
                    # Unlike gRPC, any arbitrary function (generator) cannot be passed via websockets.
                    # Simply iterating through the `req_iter` makes the request-response sequential.
                    # To make client unblocking, :func:`send_requests` and `recv_responses` are separate tasks

                    asyncio.create_task(_send_requests(request_iterator=req_iter))
                    async for response_bytes in websocket:
                        # When we have a stream of responses, instead of doing `await websocket.recv()`,
                        # we need to traverse through the websocket to recv messages.
                        # https://websockets.readthedocs.io/en/stable/faq.html#why-does-the-server-close-the-connection-after-processing-one-message

                        resp = Request(response_bytes)
                        resp = resp.as_typed_request(resp.request_type).as_response()
                        callback_exec(
                            response=resp,
                            on_error=on_error,
                            on_done=on_done,
                            on_always=on_always,
                            continue_on_error=self.args.continue_on_error,
                            logger=self.logger,
                        )
                        if self.args.show_progress:
                            p_bar.update(self.args.request_size)
                        yield resp
                        self.num_responses += 1
                        if self.num_requests == self.num_responses:
                            break

        except websockets.exceptions.ConnectionClosedOK:
            self.logger.warning(f'Client got disconnected from the websocket server')
        except websockets.exceptions.WebSocketException as e:
            self.logger.error(
                f'Got following error while streaming requests via websocket: {e!r}'
            )


class WebSocketClient(GRPCClient, WebSocketClientMixin):
    """A Python Client to stream requests from a Flow with a REST Gateway.

    :class:`WebSocketClient` shares the same interface as :class:`Client` and provides methods like
    :meth:`index`, "meth:`search`, :meth:`train`, :meth:`update` & :meth:`delete`.

    It is used by default while running operations when we create a `Flow` with `restful=True`

    .. highlight:: python
    .. code-block:: python

        from jina.flow import Flow
        f = Flow(protocol='websocket').add().add()

        with f:
            f.index(['abc'])


    :class:`WebSocketClient` can also be used to run operations for a remote Flow

    .. highlight:: python
    .. code-block:: python

        # A Flow running on remote
        from jina.flow import Flow
        f = Flow(protocol='websocket', port_expose=34567).add().add()

        with f:
            f.block()

        # Local WebSocketClient running index & search
        from jina.clients import WebSocketClient

        client = WebSocketClient(...)
        client.index(...)
        client.search(...)


    :class:`WebSocketClient` internally handles an event loop to run operations asynchronously.
    """



import uuid
from typing import Optional, BinaryIO

from jina import Flow
from jina.helper import colored, random_uuid
from .base import BaseStore
from .helper import jina_workspace
from ..models import UpdateOperationEnum


class FlowStore(BaseStore):
    """A Store of Flows for jinad"""

    def add(self, config: BinaryIO, workspace_id: Optional[uuid.UUID] = None, **kwargs):
        """Add a new Flow

        :param config: the binary data from the yaml
        :param workspace_id: the id of the workspace
        :param kwargs: other kwargs, not used
        :return: flow id"""
        try:
            if not workspace_id:
                workspace_id = random_uuid()

            flow_identity = random_uuid()
            with jina_workspace(workspace_id) as _workdir:
                y_spec = config.read().decode()
                f = Flow.load_config(y_spec)
                f.identity = str(flow_identity)
                f.workspace_id = str(workspace_id)
                f.start()

        except Exception as e:
            self._logger.error(f'{e!r}')
            raise
        else:
            self[flow_identity] = {
                'object': f,
                'arguments': vars(f.args),
                'yaml_source': y_spec,
                'workdir': _workdir,
                'workspace_id': workspace_id,
            }
            self._logger.success(
                f'{colored(str(flow_identity), "cyan")} is added to workspace {colored(str(workspace_id), "cyan")}'
            )
            return flow_identity

    def update(
        self,
        id: 'uuid.UUID',
        kind: UpdateOperationEnum,
        dump_path: str,
        pod_name: str,
        shards: int = None,
    ):
        """Run an update operation on the Flow

        :param id: the id of the Flow
        :param kind: the kind of update operation
        :param dump_path: the dump path
        :param pod_name: the pod to target
        :param shards: the nr of shards to dump for (only used for the `dump` op)"""
        flow_dict = self._items[id]
        flow_obj = flow_dict.get('object')
        ws = flow_dict.get('workspace_id')
        with jina_workspace(ws) as _workdir:
            if kind == UpdateOperationEnum.rolling_update:
                flow_obj.rolling_update(pod_name=pod_name, dump_path=dump_path)
            elif kind == UpdateOperationEnum.dump:
                raise NotImplementedError(
                    f' sending post request does not work because asyncio loop is occupied'
                )
